/********************************************************************

    backend.h

********************************************************************/


#ifndef backend_h
#define backend_h

#include "classlib/classlib.h"

#include "imcode.h"
#include "type.h"
#include "data.h"
#include "procedure.h"


const long VAR_CAPACITY = 0x10000L;
const long CODE_CAPACITY = 0x10000L;
const long CONST_CAPACITY = 0x10000L;
const int  SIZEOF_THREAD = 20;
const int  SIZEOF_THREAD_INIT = 8;


//-------------
  class BackEnd
//-------------
{
  protected:

    struct CodeWithDebugInfo
    {
      long instructionNo;
      unsigned short vmcode;  
    };

    unsigned short constants[CONST_CAPACITY];
    CodeWithDebugInfo codes[CODE_CAPACITY];

    int numThreads;
    long varBase;
    long varOffset;
    long constOffset;
    long codeOffset;
    long instructionNo;

    void alignVar ();
    
    std::stack<NumericType::FORM> types;
    
    unsigned short getThreadID ( IntermediateCode& imcode );
    int getCalcTypeID ();

    std::map<std::string, unsigned short> links;
    std::map<unsigned short, std::string> calls;
    std::map<unsigned , unsigned short>   marks;
    std::map<unsigned short, unsigned>    jumps;

    void fillInCallOffsets ();
    void fillInBranchOffsets ();

    void putCode ( unsigned short code );

    void put_LOAD_code ( IntermediateCode& imcode );
    void put_R_LOAD_code ( IntermediateCode& imcode );
    void put_ILOAD_code ( long value );
    void put_ILOAD_F_code ( double value );
    void put_INC_LOAD_code ( IntermediateCode& imcode );
    void put_STORE_code ( IntermediateCode& imcode );
    void put_R_STORE_code ( IntermediateCode& imcode );
    void put_CALL_code ( IntermediateCode& imcode );
    void put_RETURN_code ( IntermediateCode& imcode );
    void put_WIPESTACK_code ( IntermediateCode& imcode );
    void put_LOAD_PTR_code ( IntermediateCode& imcode );

    void put_STR_PUTISTR_code ( std::string& string );
    void put_STR_PUTCSTR_code ( IntermediateCode& imcode );

    void putBranchOffset ( IntermediateCode& imcode );  
    void putThreadID ( IntermediateCode& imcode );  

    void putCalcCode_A ( unsigned short code );  // add, sub ...
    void putCalcCode_B ( unsigned short code );  // equal, higher ...
    void putCalcCode_C ( unsigned short code );  // and, or ...

    void put_CAST_code ( IntermediateCode& imcode );
    void put_CAST_code ( NumericType::FORM typeID );

    void put_NEG_code ();
    void put_NOT_code ();

    unsigned short storeConst ( unsigned char b );
    unsigned short storeConst ( short i );
    unsigned short storeConst ( long l );
    unsigned short storeConst ( double f );

    void put_STR_CLEAR ();
    void put_STR_PUTCH ();
    void put_STR_PUTSTR ();
    void put_STR_PUTISTR ( IntermediateCode& imcode );
    void put_STR_PUTCSTR ();
    void put_STR_PUTVAL ();


  public:

    void init ( int _numThreads ); 

    void allocateVariable ( Variable* variable );

    unsigned short storeConst ( Constant* constant );
    unsigned short storeConst ( ArrayConstant* arrayconst );
    unsigned short storeConst ( StringConstant* strconst );
    unsigned short storeConst ( StringArrayConstant* strarrayconst );

    int  compile ( IntermediateCodeSequence* imcodeseq, const std::string& name );
    void finish ();
    void setThreadInitData ( int iThread, unsigned short codeOffset, 
                             unsigned short memoryRequirements, 
                             unsigned short localVarRequirements,
                             unsigned char priority ); 

    int getVarBase ();

    long getNumConstantBytes ();
    long getNumCodeWords ();
    long getNumVariableBytes ();

    long getConstantMemorySize ();
    long getDataMemorySize ();
    long getCodeMemorySize ();

    long getCodeOffset ();
    unsigned char  getConstByte ( int i );
    unsigned short getCodeWord ( int i );
    long getCodeInstructionNo ( int i );
};


inline long BackEnd::getCodeOffset ()
{ return codeOffset; }

inline int BackEnd::getVarBase ()
{ return varBase; }

inline long BackEnd::getNumConstantBytes ()
{ return 2 * constOffset; }

inline long BackEnd::getNumCodeWords ()
{ return codeOffset; } 

inline long BackEnd::getNumVariableBytes ()
{ return varOffset; }


inline long BackEnd::getConstantMemorySize ()
{ return CONST_CAPACITY * 2; }

inline long BackEnd::getDataMemorySize ()
{ return VAR_CAPACITY; }

inline long BackEnd::getCodeMemorySize ()
{ return CODE_CAPACITY; }


inline unsigned char BackEnd::getConstByte ( int i )
{ return *(unsigned char*)((unsigned char*)constants + i); }

inline unsigned short BackEnd::getCodeWord ( int i )
{ return codes[i].vmcode; }

inline long BackEnd::getCodeInstructionNo ( int i )
{ return codes[i].instructionNo; }


#endif