/********************************************************************

    project C-Control II Compiler

    data.cpp

********************************************************************/

    
#include "data.h"   
#include "program.h"


//---------------------------------------------------------------------------------
  Data::Data ( MEMCLASS _memClass, bool _isDataArray , unsigned _numArrayElements )
//---------------------------------------------------------------------------------
{
  memClass = _memClass;
  isDataArray = _isDataArray;
  numArrayElements = _numArrayElements;
  offset = 0;
}


//-----------------------------------
  void Data::dump( std::ostream& os )
//-----------------------------------
{
  Identified::dump(os);

  os << "\t(" << offset << ')';
}

//--- Constant ------------------------------------------------------

//---------------------
  Constant::Constant ()
//---------------------
: Data ( Data::CONSTANT )
{ 
}


//-----------------------------------------
  void Constant::setValue ( double _value )
//-----------------------------------------
{
  value = _value;
  type.updateForm(value);
}


//----------------------------
  double Constant::getValue ()
//----------------------------
{ 
  return value; 
}


//--------------------------
  Type* Constant::getType ()
//--------------------------
{
  return &type;
}


//-----------------------------
  unsigned Constant::getSize ()
//-----------------------------
{
  return type.getSize();
}


//--- ArrayConstant -------------------------------------------------

//-------------------------------
  ArrayConstant::ArrayConstant ()
//-------------------------------
: Data ( Data::CONSTANT, true )
{ 
  type.updateForm(-1);  // at least INT!
}


//----------------------------------------------
  void ArrayConstant::addValue ( double _value )
//----------------------------------------------
{
  value.push_back(_value);
  type.updateForm(_value);
  numArrayElements++;
}


//---------------------------------------------
  double ArrayConstant::getValue ( unsigned i )
//---------------------------------------------
{ 
  return value[i]; 
}


//-------------------------------
  Type* ArrayConstant::getType ()
//-------------------------------
{
  return &type;
}


//----------------------------------
  unsigned ArrayConstant::getSize ()
//----------------------------------
{
  return type.getSize() * numArrayElements;
}


//--- StringConstant ------------------------------------------------


//---------------------------------
  StringConstant::StringConstant ()
//---------------------------------
: Data ( Data::CONSTANT )
{
}


//-------------------------------------------------------------
  void StringConstant::setString ( const std::string& _string )
//-------------------------------------------------------------
{
  string = _string;
}


//-----------------------------------------------
  const std::string& StringConstant::getString ()
//-----------------------------------------------
{
  return string;
}


//--------------------------------
  Type* StringConstant::getType ()
//--------------------------------
{
  return &StandardType::STRING;
}


//-----------------------------------
  unsigned StringConstant::getSize ()
//-----------------------------------
{
  return string.size(); 
}


//--- StringArrayConstant -------------------------------------------


//-------------------------------------------
  StringArrayConstant::StringArrayConstant ()
//-------------------------------------------
: Data ( Data::CONSTANT, true )
{
}


//------------------------------------------------------------------
  void StringArrayConstant::addString ( const std::string& _string )
//------------------------------------------------------------------
{
  string.push_back(_string);
  numArrayElements++;
}


//----------------------------------------------------------------
  const std::string& StringArrayConstant::getString ( unsigned i )
//----------------------------------------------------------------
{
  return string[i];
}


//-------------------------------------
  Type* StringArrayConstant::getType ()
//-------------------------------------
{
  return &StandardType::STRING;
}


//----------------------------------------
  unsigned StringArrayConstant::getSize ()
//----------------------------------------
{
  unsigned numStrings = string.size();
  unsigned cumulativeSize = 0;

  for ( unsigned i = 0; i < numStrings; i++ )
  {
    cumulativeSize += string[i].size();
  }

  return numStrings * sizeof (unsigned short) + cumulativeSize; 
}




//--- Reference ----------------------------------------------------

//--------------------------------------------------------
  Reference::Reference ( Type* _type, bool _isDataArray  )
//--------------------------------------------------------
: Data(Data::LOCAL, _isDataArray)
{
  type = _type;
}


//---------------------------
  Type* Reference::getType ()
//---------------------------
{
  return type;
}


//------------------------------
  unsigned Reference::getSize ()
//------------------------------
{
  return Data::PTR_SIZE;
}


//------------------------------------------------------------------------
  Variable::Variable ( Type* _type, MEMCLASS _memClass, bool _isDataArray, 
                       unsigned _numArrayElements, Reference* _reference )
//------------------------------------------------------------------------
: Data (_memClass, _isDataArray, _numArrayElements)
{
  type = _type;
  reference = _reference;
}


//-----------------------------------
  int Variable::getReferenceOffset ()
//-----------------------------------
{
  return reference->getOffset();
}
    
    
//--------------------------
  Type* Variable::getType ()
//--------------------------
{
  return type;
}


//-----------------------------
  unsigned Variable::getSize () 
//-----------------------------
{
  return type->getSize() * (isArray() ? numArrayElements : 1);
}


//---------------------
  void Variable::use ()
//---------------------
{
  Identified::use();

  Identifier* parentIdent = identifier->getParent();
  Identified* identified = dynamic_cast<Identified*>(parentIdent->getValue().get());

  if ( identified->what() == VAR_IDENT )
    identified->use();
}
