/********************************************************************

    project C-Control II Compiler

    data.h

********************************************************************/

#ifndef data_h
#define data_h


#include "classlib/classlib.h"
#include "identifier.h"
#include "type.h"


//------------------------------
  class Data : public Identified
//------------------------------
{
  public: 

    typedef enum { CONSTANT, STATIC, LOCAL } MEMCLASS;
    enum { PTR_SIZE = 2 };

  protected:

    MEMCLASS memClass;
    int      offset;
    bool     isDataArray;
    unsigned numArrayElements;          

    Data ();

  public:

    Data ( MEMCLASS _memClass, bool _isDataArray = false, unsigned _numArrayElements = 0 );

    MEMCLASS getMemoryClass ();
    bool isArray ();
    unsigned getNumArrayElements ();

    void setOffset ( int _offset );
    int  getOffset ();

    virtual Type* getType () = 0;
    virtual unsigned getSize () = 0; 

    bool isUserType (); 

    virtual void dump( std::ostream& os );
};


inline Data::MEMCLASS Data::getMemoryClass ()
{ return memClass; }

inline bool Data::isArray ()
{ return isDataArray; }

inline unsigned Data::getNumArrayElements ()
{ return numArrayElements; }

inline void Data::setOffset ( int _offset )
{ offset = _offset; }

inline int Data::getOffset ()
{ return offset; }

inline bool Data::isUserType ()
{ return getType()->isUserType(); }


//----------------------------
  class Constant : public Data
//----------------------------
{
  protected:

    double value;
    NumericType type;
 
  public:

    Constant();

    void   setValue ( double _value );
    double getValue ();

    virtual Type* getType ();
    virtual unsigned getSize (); 
};


//---------------------------------
  class ArrayConstant : public Data
//---------------------------------
{
  protected:

    std::vector<double> value;
    NumericType type;

  public:

    ArrayConstant ();

    void   addValue ( double _value );
    double getValue ( unsigned i );

    virtual Type* getType ();
    virtual unsigned getSize (); 
};


//----------------------------------
  class StringConstant : public Data
//----------------------------------
{
  protected:
    
    std::string string;

  public:
    
    StringConstant ();

    void setString ( const std::string& _string );
    const std::string& getString ();

    virtual Type* getType ();
    virtual unsigned getSize (); 
};


//---------------------------------------
  class StringArrayConstant : public Data
//---------------------------------------
{
  protected:

    std::vector<std::string> string;

  public:
    
    StringArrayConstant ();

    void addString ( const std::string& _string );
    const std::string& getString ( unsigned i );

    virtual Type* getType ();
    virtual unsigned getSize (); 
};


//-----------------------------
  class Reference : public Data
//-----------------------------
{
  protected:

    Type* type;

    Reference ();

  public:

    Reference ( Type* _type, bool _isDataArray = false );

    virtual Type* getType ();
    virtual unsigned getSize (); 
};


//----------------------------
  class Variable : public Data
//----------------------------
{
  protected:

    Type* type;
    Reference* reference;      // for indirect access

    Variable ();

  public:

    Variable ( Type* _type, MEMCLASS _memClass, bool _isDataArray = false, 
               unsigned _numArrayElements = 0, Reference* _reference = 0 );

    int getReferenceOffset ();

    virtual Type* getType ();
    virtual unsigned getSize (); 

    Reference* getReference () const;

    bool isIndirectVariable () const;

    virtual void use ();
};


inline bool Variable::isIndirectVariable () const
{ return 0 != reference; }

inline Reference* Variable::getReference () const
{ return reference; }

  
#endif