/********************************************************************

    cc2comp.cpp : Definiert den Einsprungpunkt fr die DLL-Anwendung.
//

********************************************************************/


#include "cc2comp.h"

#ifdef _MSC_VER
  #pragma warning(disable:4786)
#endif

#include "../project.h"



//-------------------------------------------------------------------------------------
  BOOL APIENTRY DllMain( HANDLE hModule, DWORD  ul_reason_for_call, LPVOID lpReserved	)
//-------------------------------------------------------------------------------------
{
  switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
  }

  return TRUE;
}


//--- exported functions --------------------------------------------

//----------------------------------------------
  PROJECT CC2COMP_API openProject ( LPSTR name )
//----------------------------------------------
{
  try
  {
    if ( name )
    {
      PROJECT project = static_cast<PROJECT>(new Project(std::string(name)));
      return project;
    }
    else
      return 0;
  }
  catch (...)
  {
    return 0;
  }
}


//-------------------------------------------------
  BOOL CC2COMP_API closeProject ( PROJECT project )
//-------------------------------------------------
{
  try
  {
    delete static_cast<Project*>(project);
    return TRUE;
  }
  catch (...)
  {
    return FALSE;
  }
}


//--------------------------------------------------------------
  BOOL CC2COMP_API renameProject ( PROJECT project, LPSTR name )
//--------------------------------------------------------------
{
  try
  {
    Project* prj = static_cast<Project*>(project);
    prj->rename(name);
    return TRUE;
  }
  catch (...)
  {
    return FALSE;
  }
}


//-------------------------------------------------
  BOOL CC2COMP_API isValidModuleName( LPCSTR name )
//-------------------------------------------------
{
  static Scanner scanner;

  std::istrstream input(name);
  ScanResult output;

  scanner.run(&input, &output);
  output.gotoToken(0);

  return ( output.getNumberOfTokens() == 2 ) && ( output.getToken().id == Token::IDENTIFIER );
}


//---------------------------------------------------------------------------------------------------
  C2_MODULE_HANDLE CC2COMP_API addModule ( PROJECT project, LPSTR name, LPSTR source, DWORD options )
//---------------------------------------------------------------------------------------------------
{
  try
  {
    bool isFileSource = (options == 1);
    std::string filename = isFileSource ? source : "?";
    char* strbuf = isFileSource ? 0 : source;
    
    return static_cast<Project*>(project)->add(std::string(name), filename, strbuf);
  }
  catch (...)
  {
    return 0;
  }
}


//--------------------------------------------------------------------------------------------------------------  
  BOOL CC2COMP_API replaceModuleSource ( PROJECT project, C2_MODULE_HANDLE module, LPSTR source, DWORD options )
//--------------------------------------------------------------------------------------------------------------  
{
  try
  {
    Module* ptr = static_cast<Project*>(project)->getModule(module);
    
    if ( options == 1)
      ptr->setSource(std::string(source));
    else
      ptr->setSource(source);

    return TRUE;
  }
  catch (...)
  {
    return FALSE;
  }
}


//----------------------------------------------
  BOOL CC2COMP_API removeAll ( PROJECT project )
//----------------------------------------------
{
  try
  {
    static_cast<Project*>(project)->removeAll();
    return TRUE;
  }
  catch (...)
  {
    return FALSE;
  }
}


//--------------------------------------------------------------------------
  BOOL CC2COMP_API removeModule ( PROJECT project, C2_MODULE_HANDLE module )   
//--------------------------------------------------------------------------
{
  try
  {
    return static_cast<Project*>(project)->remove(module);
  }
  catch (...)
  {
    return FALSE;
  }
}


//---------------------------------------------------------------------------------
  BOOL CC2COMP_API setMessageBuffer ( PROJECT project, LPSTR output, DWORD length )
//---------------------------------------------------------------------------------
{
  try
  {
    static_cast<Project*>(project)->setErrLog(output, length);
    return TRUE;
  }
  catch (...)
  {
    return FALSE;
  }
}


//----------------------------------------------------------------------------
  DWORD CC2COMP_API compileModule ( PROJECT project, C2_MODULE_HANDLE module )
//----------------------------------------------------------------------------
{
  try
  {
    long result = static_cast<Project*>(project)->compile(module);

    if ( result < 0 )
      return 0;
    else if ( result == 0 )
      return 0xFFFFFFFFL;
    else 
      return static_cast<DWORD>(result);     
  }
  catch (...)
  {
    return 0;
  }
}


//-----------------------------------------------
  BOOL CC2COMP_API beginBuild ( PROJECT project )
//-----------------------------------------------
{
  try
  {
    Project* prj = static_cast<Project*>(project);
    prj->beginBuild();

    return TRUE;
  }
  catch (...)
  {
    return FALSE;
  }
}


//------------------------------------------------
  BOOL CC2COMP_API finishBuild ( PROJECT project )
//------------------------------------------------
{
  try
  {
    Project* prj = static_cast<Project*>(project);
    return prj->finishBuild()? TRUE : FALSE;
  }
  catch (...)
  {
    return FALSE;
  }
}



//-------------------------------------------------------------------------------
  BOOL CC2COMP_API getModuleLineInfos ( PROJECT project, C2_MODULE_HANDLE module, 
                                        LPSTR infos, DWORD linecount )
//-------------------------------------------------------------------------------
{
  try
  {
    Project* prj = static_cast<Project*>(project);
    Module* mod  = prj->getModule(module);
    Module::LineInfos* lineinfos = mod->getLineInfos();

    for ( DWORD i = 0; (i < linecount) && (i < lineinfos->size()); i++ )
      infos[i] = (*lineinfos)[i];

    return TRUE;
  }
  catch (...)
  {
    return FALSE;
  }
}


//---------------------------------------------------------------------------------
  BOOL CC2COMP_API getIdentDefPos ( PROJECT project, LPSTR ident, 
                                    C2_MODULE_HANDLE* module_ptr, DWORD* lino_ptr )
//---------------------------------------------------------------------------------
{
  try
  {
    Project* prj = static_cast<Project*>(project);

    if ( prj->getIdentDefPos(ident, module_ptr, lino_ptr) )
      return TRUE;
    else
      return FALSE;
  }
  catch (...)
  {
    return FALSE;
  }
}


//---------------------------------------------------------------------------
  BOOL CC2COMP_API getIdentCompletions ( PROJECT project, LPSTR ident, 
                                         C2_MODULE_HANDLE module, DWORD lino, 
                                         LPSTR completions, DWORD length )
//---------------------------------------------------------------------------
{
  try
  {
    Project* prj = static_cast<Project*>(project);

    if ( prj->getIdentCompletions(ident, module, lino, completions, length) )
      return TRUE;
    else
      return FALSE;
  }
  catch (...)
  {
    return FALSE;
  }
}


//------------------------------------------------------------------------
  BOOL CC2COMP_API getIdentFullName ( PROJECT project, LPSTR ident, 
                                      C2_MODULE_HANDLE module, DWORD lino, 
                                      LPSTR fullname, DWORD length )
//------------------------------------------------------------------------
{
  try
  {
    std::ostrstream os(fullname, length-1);

    Project* prj = static_cast<Project*>(project);

    Identifier* identifier = prj->getIdentifierInContext(ident, module, lino);
    identifier->writePath(os, '.');
    os << '\0';

    return TRUE;
  }
  catch (...)
  {
    return FALSE;
  }
}


//------------------------------------------------------------------
  BOOL CC2COMP_API writeOutputFile ( PROJECT project, LPSTR output )
//------------------------------------------------------------------
{
  try
  {
    Project* prj = static_cast<Project*>(project);
    std::string outputFileName = output;
    prj->writeOutputFile(outputFileName);
    
    return TRUE;
  }
  catch (...)
  {
    return FALSE;
  }
}


//--------------------------------------------------------------
  DWORD CC2COMP_API getRequiredVMCBufferSize ( PROJECT project )
//--------------------------------------------------------------
{
  try
  {
    Project* prj = static_cast<Project*>(project);
    return prj->getRequiredVMCBufferSize();
  }
  catch (...)
  {
    return 0;
  }
}


//----------------------------------------------------------------
  BOOL CC2COMP_API fillVMCBuffer ( PROJECT project, LPSTR buffer )
//----------------------------------------------------------------
{
  try
  {
    Project* prj = static_cast<Project*>(project);
    prj->fillVMCBuffer((unsigned char*)buffer);

    return TRUE;
  }
  catch (...)
  {
    return FALSE;
  }
}


