/********************************************************************

    project C-Control II Compiler

    imcode.cpp

********************************************************************/


#include "imcode.h"
#include "text.h"
#include "data.h"
#include "procedure.h"
#include "error.h"



/*
static const char* optxt[] =
{
  "NOP", "CALL", "RETURN", "WIPESTACK", "-----", "GOTO", "BRANCH_IF_ZERO", 
  "CAST", "LOAD", "I_LOAD", "I_LOAD_F", "LOAD_PTR", "R_LOAD", "INC_LOAD", "R_INC_LOAD", 
  "STORE", "R_STORE",	"K_STORE",
  "NEG", "INVERT",
  "ADD", "SUB", "MUL", "DIV", "MOD", 
  "EQUAL", "NOTEQUAL", "HIGHER", "LOWER", "HIGHERSAME", "LOWERSAME",
  "AND", "NAND", "OR", "NOR", "XOR", "NXOR", 
  "SHL", "SHR", "ASR", "ROL", "ROR",

  "STR_CLEAR", "STR_PUTCH", "STR_PUTSTR", "STR_PUTISTR", "STR_PUTCSTR", "STR_PUTVAL",

  "INLINE", "RUN", "HALT", "RESUME", "RESET",
  "SLEEP", "CAPTURE", "RELEASE", "YIELD", "QUIT",
  "ENTER_INSTRUCTION"
};


static const char* typetxt[] =
{ "BYTE", "INT", "LONG", "FLOAT", "USERTYPE" };


//------------------------------------------------
  void IntermediateCode::dump ( std::ostream& os )
//------------------------------------------------
{
  os << optxt[opcode];
}

//--------------------------------------------
  void ImCodeWithID::dump ( std::ostream& os )
//--------------------------------------------
{
  IntermediateCode::dump(os);
  os << '\t';

  switch ( opcode )
  {
    case IntermediateCode::CAST: 

      os << typetxt[id]; 
      break;

    case IntermediateCode::MARK:
    case IntermediateCode::GOTO:
    case IntermediateCode::BRANCH_IF_ZERO: 

      os << '#';

    default : os << id; break;
  }
}


//--------------------------------------------------
  void ImCodeImmLoadValue::dump ( std::ostream& os )
//--------------------------------------------------
{
  IntermediateCode::dump(os);
  os << '\t' << value;
}


//-------------------------------------------------------
  void ImCodeImmLoadFloatValue::dump ( std::ostream& os )
//-------------------------------------------------------
{
  IntermediateCode::dump(os);
  os << '\t' << value;
}


//---------------------------------------
  void ImCodeC::dump ( std::ostream& os )
//---------------------------------------
{
  IntermediateCode::dump(os);
  os << '\t' << value;
}

//---------------------------------------
  void ImCodeD::dump ( std::ostream& os )
//---------------------------------------
{
  IntermediateCode::dump(os);
  os << '\t';

  if ( object ) 
    object->dump(os);
}*/



//--- IntermediateCode ----------------------------------------------


//------------------------------------------------
  bool IntermediateCode::hasConstantValue () const
//------------------------------------------------
{ 
  if ( opcode == I_LOAD ) return true;
  if ( opcode == I_LOAD_F ) return true;

  if ( opcode == LOAD ) 
    if ( typeid(*enclosed.object) == typeid(Constant) )  
      return true;

  return false;
}



//--- IntermediateCodeSequence --------------------------------------

unsigned IntermediateCodeSequence::nextmark = 0;


//-----------------------------------------------------
  IntermediateCodeSequence::IntermediateCodeSequence ()
//-----------------------------------------------------
{ 
  codes = 0;
  floats = 0;
  strings = 0;
  constCounter = 0;

  init();

  localMemSize = 0; 
  ownMemoryRequirements = 0;
  callMemoryRequirements = 0;
}


//------------------------------------------------------
  IntermediateCodeSequence::~IntermediateCodeSequence ()
//------------------------------------------------------
{ 
  free();
}


//--------------------------------------
  void IntermediateCodeSequence::init ()
//--------------------------------------
{
  free();

  codeCapacity   = 1024;
  floatCapacity  = 4;
  stringCapacity = 1;

  codes = new IntermediateCode[codeCapacity];
  floats = new double[floatCapacity];
  strings = new std::string[stringCapacity];
}


//--------------------------------------
  void IntermediateCodeSequence::free ()
//--------------------------------------
{
  delete [] codes;
  delete [] floats;
  delete [] strings;
  
  codes = 0;
  floats = 0;
  strings = 0;

  codeCapacity = 0;
  floatCapacity = 0;
  stringCapacity = 0;

  numberOfCodes = 0;
  numberOfFloats = 0;
  numberOfStrings = 0;

  iCode = 0;
  iFloat = 0;
  iString = 0;
}


//----------------------------------------
  void IntermediateCodeSequence::finish ()
//----------------------------------------
{
  if ( !numberOfCodes )
  { 
    numberOfCodes   = iCode;
    numberOfStrings = iString;
    numberOfFloats  = iFloat;
  }
}


//------------------------------------------------------------------------------
  void IntermediateCodeSequence::pushCode ( IntermediateCode::OPERATION opcode )
//------------------------------------------------------------------------------
{
  if ( iCode == codeCapacity )
  {
    unsigned long newCapacity = codeCapacity * 2;
    IntermediateCode* newPointer = new IntermediateCode[newCapacity];

    while ( codeCapacity-- )
      newPointer[codeCapacity] = codes[codeCapacity];

    delete [] codes;
    codes = newPointer;
    codeCapacity = newCapacity;
  }

  codes[iCode++].opcode = opcode;
}


//------------------------------------------------------------------------------------------
  void IntermediateCodeSequence::pushCode ( IntermediateCode::OPERATION opcode, long value )
//------------------------------------------------------------------------------------------
{
  pushCode(opcode);
  codes[iCode-1].enclosed.value = value;
}


//-----------------------------------------------------------------------------------------------
  void IntermediateCodeSequence::pushCode ( IntermediateCode::OPERATION opcode, Virtual* object )
//-----------------------------------------------------------------------------------------------
{
  pushCode(opcode);
  codes[iCode-1].enclosed.object = object;
}


//-------------------------------------------------------------------------------------------------
  void IntermediateCodeSequence::pushFloatCode ( IntermediateCode::OPERATION opcode, double value )
//-------------------------------------------------------------------------------------------------
{
  pushCode(opcode);
  codes[iCode-1].enclosed.value = iFloat;

  if ( iFloat == floatCapacity )
  {
    unsigned long newCapacity = floatCapacity * 2;
    double* newPointer = new double[newCapacity];

    while ( floatCapacity-- )
      newPointer[floatCapacity] = floats[floatCapacity];

    delete [] floats;
    floats = newPointer;
    floatCapacity = newCapacity;
  }
  floats[iFloat++] = value;
}


//-----------------------------------------------------------------------------------
  void IntermediateCodeSequence::pushStringCode ( IntermediateCode::OPERATION opcode, 
                                                  const std::string& string )
//-----------------------------------------------------------------------------------
{
  pushCode(opcode);
  codes[iCode-1].enclosed.value = iString;

  if ( iString == stringCapacity )
  {
    unsigned long newCapacity = stringCapacity * 2;
    std::string* newPointer = new std::string[newCapacity];

    while ( stringCapacity-- )
      newPointer[stringCapacity] = strings[stringCapacity];

    delete [] strings;
    strings = newPointer;
    stringCapacity = newCapacity;
  }
  strings[iString++] = string;
}


//-----------------------------------------------------------------------------------
  void IntermediateCodeSequence::chainCodeSeq ( IntermediateCodeSequence* seq, 
                                                bool shallCopyInstructionChangeInfo )
//-----------------------------------------------------------------------------------
{
  for ( seq->rewind(); !seq->isLastCode(); seq->gotoNextCode() )
  {
    switch ( seq->getCode().opcode ) 
    {
      case IntermediateCode::I_LOAD_F : 
      
        pushFloatCode(IntermediateCode::I_LOAD_F, seq->getCodeEnclosedFloat());
        break;

      case IntermediateCode::STR_PUTISTR : 
        
        pushStringCode(IntermediateCode::STR_PUTISTR, seq->getCodeEnclosedString()); 
        break;

      case IntermediateCode::ENTER_INSTRUCTION:
        
        if ( !shallCopyInstructionChangeInfo )
          continue;

      default: 
        
        pushCode(seq->getCode().opcode);
        codesBack().enclosed = seq->getCode().enclosed;
        break;
    }
  }
}


//-----------------------------------------------------------------------------
  void IntermediateCodeSequence::putCode ( IntermediateCode::OPERATION opcode )
//-----------------------------------------------------------------------------
{ 
  if ( ! (constCounter && IntermediateCode::isNumericalOperation(opcode))  )
  {
    constCounter = 0;
    pushCode(opcode); 
    return;
  }

  // *** fold constant sub-expressions ***

  if ( constCounter == 1 )
  {
    double value = getConstantResult();

    switch ( opcode )
    {
      case IntermediateCode::NEG : 
      
        popCode();

        if ( value == (long)value ) 
          pushCode(IntermediateCode::I_LOAD, -(long)value); 
        else
          pushFloatCode(IntermediateCode::I_LOAD_F, -value); 

        break;
  
      case IntermediateCode::NOT : 
         
        popCode();

        if ( value == (long)value ) 
          pushCode(IntermediateCode::I_LOAD, !(long)value); 
        else
          pushFloatCode(IntermediateCode::I_LOAD_F, ~(long)value); 

        break;
      
      default:  
           
        constCounter = 0;
        pushCode(opcode);
        break;
    }
  }
  else
  {
    double b = getConstantResult();
    popCode();
    double a = getConstantResult();
    popCode();
    
    switch ( opcode )
    {
      case IntermediateCode::ADD        : a += b; break; 
      case IntermediateCode::SUB        : a -= b; break; 
      case IntermediateCode::MUL        : a *= b; break; 
      case IntermediateCode::DIV        : if (0 == b) throw Error(); a /= b; break; 
      case IntermediateCode::MOD        : if (0 == b) throw Error(); a = (long) a % (long) b; break; 
      case IntermediateCode::SHL        : a = (unsigned long) a << (unsigned long) b; break; 
      case IntermediateCode::SHR        : a = (unsigned long) a >> (unsigned long) b; break; 
      case IntermediateCode::AND        : a = (unsigned long) a & (unsigned long) b; break; 
      case IntermediateCode::NAND       : a = !((unsigned long) a & (unsigned long) b); break; 
      case IntermediateCode::OR         : a = (unsigned long) a | (unsigned long) b; break; 
      case IntermediateCode::NOR        : a = !((unsigned long) a | (unsigned long) b); break; 
      case IntermediateCode::XOR        : a = (unsigned long) a ^ (unsigned long) b; break; 
      case IntermediateCode::EQUAL      : a = (a == b) ? -1 : 0; break; 
      case IntermediateCode::NOTEQUAL   : a = (a != b) ? -1 : 0; break; 
      case IntermediateCode::HIGHER     : a = (a > b)  ? -1 : 0; break; 
      case IntermediateCode::LOWER      : a = (a < b)  ? -1 : 0; break; 
      case IntermediateCode::HIGHERSAME : a = (a >= b) ? -1 : 0; break; 
      case IntermediateCode::LOWERSAME  : a = (a <= b) ? -1 : 0; break; 

      default : break;
    }

    constCounter--;

    if ( a == (long)a ) 
      pushCode(IntermediateCode::I_LOAD, (long)a); 
    else
      pushFloatCode(IntermediateCode::I_LOAD_F, a); 
  }
}


//----------------------------------------------------------------------------------
  void IntermediateCodeSequence::putCodeWithID ( IntermediateCode::OPERATION opcode, 
                                                 unsigned id )
//----------------------------------------------------------------------------------
{ 
  pushCode(opcode, (long)id); 
}


/*
//-------------------------------------------------------------------------------------------
  void IntermediateCodeSequence::putCode ( IntermediateCode::OPERATION opcode, double value )
//-------------------------------------------------------------------------------------------
{ 
  if ( IntermediateCode::I_LOAD == opcode )
    constCounter++;
  else
    constCounter = 0;  

  codes.push_back(IntermediateCode::Pointer(new ImCodeC(opcode, value))); 
}
*/


//-----------------------------------------------------------------
  void IntermediateCodeSequence::putImmLoadValueCode ( long value )
//-----------------------------------------------------------------
{
  constCounter++;
  pushCode(IntermediateCode::I_LOAD, value); 
}


//------------------------------------------------------------------------
  void IntermediateCodeSequence::putImmLoadFloatValueCode ( double value )
//------------------------------------------------------------------------
{
  constCounter++;
  pushFloatCode(IntermediateCode::I_LOAD_F, value); 
}


//--------------------------------------------------------------------------------------
  void IntermediateCodeSequence::putCodeWithObject ( IntermediateCode::OPERATION opcode, 
                                                     Virtual* object )
//--------------------------------------------------------------------------------------
{ 
  unsigned _constCounter = constCounter;
  constCounter = 0;

  if ( iCode )  // codes not empty
    if ( (IntermediateCode::RETURN == opcode) && (IntermediateCode::RETURN == codesBack().opcode) )
      return;      

  if ( IntermediateCode::LOAD == opcode ) 
  {
    if ( typeid(*object) == typeid(Constant) )
      constCounter = _constCounter + 1;
      
    if ( iCode ) // codes not empty
    {
      if ( IntermediateCode::STORE == codesBack().opcode )
      {
        Variable* var = dynamic_cast<Variable*>(codesBack().getObject());

        if ( (object == codesBack().getObject()) && !var->isIndirectVariable() )
        {
          popCode();
          pushCode(IntermediateCode::K_STORE, object);
          return;      
        }
      }
    }
  }

  pushCode(opcode, object);
}


//--------------------------------------------------------------------------------------
  void IntermediateCodeSequence::putCodeWithString ( IntermediateCode::OPERATION opcode, 
                                                     const std::string& string )
//--------------------------------------------------------------------------------------
{
  pushStringCode(opcode, const_cast<std::string&>(string));
}




//---------------------------------------------
  unsigned IntermediateCodeSequence::getMark ()
//---------------------------------------------
{ 
  return ++nextmark; 
}


//------------------------------------------
  unsigned IntermediateCodeSequence::mark ()
//------------------------------------------
{ 
  putCodeWithID(IntermediateCode::MARK, getMark());
  return nextmark; 
}


//---------------------------------------------------
  void IntermediateCodeSequence::mark ( unsigned id )
//---------------------------------------------------
{ 
  putCodeWithID(IntermediateCode::MARK, id);
}


//---------------------------------------------------------------
  void IntermediateCodeSequence::requireMemory ( unsigned  size )
//---------------------------------------------------------------
{ 
  ownMemoryRequirements += size; 
} 


//----------------------------------------------------------------------
  void IntermediateCodeSequence::requireMemoryForCall ( unsigned  size )
//----------------------------------------------------------------------
{ 
  if ( size > callMemoryRequirements ) 
    callMemoryRequirements = size; 
} 


//-----------------------------------------------------------
  unsigned IntermediateCodeSequence::getMemoryRequirements ()
//-----------------------------------------------------------
{ 
  return ownMemoryRequirements + callMemoryRequirements; 
} 


//-----------------------------------------------------
  double IntermediateCodeSequence::getConstantResult ()
//-----------------------------------------------------
{
  if ( !codesBack().hasConstantValue() ) 
    throw NO_CONST(); 

  if ( codesBack().opcode == IntermediateCode::I_LOAD )
    return codesBack().getValue();

  if ( codesBack().opcode == IntermediateCode::I_LOAD_F )
    return floats[codesBack().getValue()];

  if ( codesBack().opcode == IntermediateCode::LOAD )
  { 
    Constant* constant = dynamic_cast<Constant*>(codesBack().getObject());
    return constant->getValue();
  }

  return 0;
}


/*
//--------------------------------------------------------  
  void IntermediateCodeSequence::dump ( std::ostream& os )
//--------------------------------------------------------  
{
  IntermediateCode::List::iterator imCodeIter;

  for ( imCodeIter = codes.begin(); imCodeIter != codes.end(); ++imCodeIter )
  {
    os << "\t";
    (*imCodeIter)->dump(os); 
    os << "\n";
  }
}
*/


