/********************************************************************

    project C-Control II Compiler

    module.cpp

********************************************************************/


#include "module.h"


//----------------------------------------------------------------------
  Module::Module ( const std::string& _filename, unsigned long _handle )
//----------------------------------------------------------------------
{
  handle = _handle;
  setSource(_filename);
}


//-------------------------------------------------------
  Module::Module ( char* _strbuf, unsigned long _handle )
//-------------------------------------------------------
{
  handle = _handle;
  setSource(_strbuf);
}


//-------------------------------------------------------
  void Module::setSource ( const std::string& _filename )
//-------------------------------------------------------
{
  filename = _filename;
  strbuf = 0;
}


//----------------------------------------
  void Module::setSource ( char* _strbuf )
//----------------------------------------
{
  filename = "?";
  strbuf = _strbuf;
}


//-------------------------------------------
  const std::string& Module::getSourceName ()
//-------------------------------------------
{
  if ( strbuf )
    return getName();
  else 
    return filename;
}


//---------------------
  void Module::reset ()
//---------------------
{
  tokens.free();
  lineinfos.clear();
  contextlist.clear();
  identifier->removeChildren();
}


//---------------------------
  void Module::clearTokens ()
//---------------------------
{
  tokens.free();
}



//--------------------------
  void Module::openSource ()
//--------------------------
{
   source = strbuf ? SourcePointer(new std::istrstream(strbuf)) : 
                     SourcePointer(new std::ifstream(filename.c_str()));
}


//---------------------------
  void Module::closeSource ()
//---------------------------
{
  source.release();  
}


//-------------------------------------------------------------------------
  void Module::changeContext ( unsigned long lino, Identifier* identifier )
//-------------------------------------------------------------------------
{
  Context context = {lino, identifier};
  contextlist.push_back(context);
}


//-----------------------------------------------------
  Identifier* Module::getContext ( unsigned long lino )
//-----------------------------------------------------
{
  Identifier* context = getIdentifier();

  std::list<Context>::iterator iterator;

  for ( iterator = contextlist.begin(); iterator != contextlist.end(); ++iterator )
  {
    if ( lino > (*iterator).lino )
      context = (*iterator).identifier;
    else
      break;
  }



  return context;
}