/********************************************************************

    project C-Control II Compiler

    module.h

********************************************************************/

#ifndef module_h
#define module_h

#include "classlib/classlib.h"

#include "token.h"
#include "identifier.h"
#include "scanner.h"

#include "type.h"
#include "data.h"
#include "procedure.h"


//--------------------------------
  class Module : public Identified
//--------------------------------
{
  public:

    typedef enum { LINE_EMPTY = 0, LINE_EXECUTABLE = 1, LINE_ERROR = 128 } LineInfo;
    typedef std::vector<unsigned char> LineInfos;

    struct Context { unsigned long lino; Identifier* identifier; };
    
    typedef std::list<Context> ContextList;


  protected:

    typedef std::auto_ptr<std::istream> SourcePointer;

    unsigned long handle;            
    std::string   filename;
    char*         strbuf;
    SourcePointer source;
    ScanResult    tokens;
    LineInfos     lineinfos;
    ContextList   contextlist;

    Module ();

  public:

    Module ( const std::string& _filename, unsigned long _handle );
    Module ( char* _strbuf, unsigned long _handle );

    unsigned long getHandle ();

    void setSource ( const std::string& _filename );
    void setSource ( char* _strbuf );

    const std::string& getName ();
    const std::string& getSourceName ();

    void reset ();
    void openSource ();
    void closeSource ();

    std::istream*  getSource ();
    ScanResult*    getTokens ();
    LineInfos*     getLineInfos ();
    void clearTokens ();

    void changeContext  ( unsigned long lino, Identifier* identifier );
    Identifier* getContext ( unsigned long lino );
};


inline unsigned long Module::getHandle ()
{ return handle; }

inline const std::string& Module::getName ()
{ return identifier->getKey(); }

inline std::istream* Module::getSource ()
{ return source.get(); }

inline ScanResult* Module::getTokens ()
{ return &tokens; }

inline Module::LineInfos* Module::getLineInfos ()
{ return &lineinfos; }

#endif