/********************************************************************

    project C-Control II Compiler

    parsterm.cpp

********************************************************************/

#include "parser.h"
#include "error.h"
#include "text.h"


//-------------------------
  void Parser::parseTerm ()
//-------------------------
{
  parseOr();
}


//-----------------------
  void Parser::parseOr ()
//-----------------------
{
  IntermediateCode::OPERATION opcode;

  parseAnd();

  for(;;) 
  {
    switch ( readNextToken() )
    {
      case Token::OR  : opcode = IntermediateCode::OR; break;
      case Token::NOR : opcode = IntermediateCode::NOR; break;
      case Token::XOR : opcode = IntermediateCode::XOR; break;

      default : opcode = IntermediateCode::NOP;
    }
    if ( IntermediateCode::NOP != opcode )
    {
      readNextToken();
      parseAnd();
      putCode(opcode);
    }
    else
    { 
      putBackToken();
      break;
    }
  }  
}


//------------------------
  void Parser::parseAnd ()
//------------------------
{
  IntermediateCode::OPERATION opcode;

  parseCompare();

  for(;;) 
  {
    switch ( readNextToken() )
    {
      case Token::AND : opcode = IntermediateCode::AND; break;
      case Token::NAND : opcode = IntermediateCode::NAND; break;

      default : opcode = IntermediateCode::NOP;
    }
    if ( IntermediateCode::NOP != opcode )
    {
      readNextToken();
      parseCompare();
      putCode(opcode);
    }
    else
    { 
      putBackToken();
      break;
    }
  }  
}


//----------------------------
  void Parser::parseCompare ()
//----------------------------
{
  IntermediateCode::OPERATION opcode;

  parseShift();

  for(;;) 
  {
    switch ( readNextToken() )
    {
      case Token::EQUAL      : opcode = IntermediateCode::EQUAL; break;
      case Token::NOTEQUAL   : opcode = IntermediateCode::NOTEQUAL; break;
      case Token::HIGHER     : opcode = IntermediateCode::HIGHER; break;
      case Token::LOWER      : opcode = IntermediateCode::LOWER; break;
      case Token::HIGHERSAME : opcode = IntermediateCode::HIGHERSAME; break;
      case Token::LOWERSAME  : opcode = IntermediateCode::LOWERSAME; break;

      default : opcode = IntermediateCode::NOP;
    }
    if ( IntermediateCode::NOP != opcode )
    {
      readNextToken();
      parseShift();
      putCode(opcode);
    }
    else
    { 
      putBackToken();
      break;
    }
  }  
}


//--------------------------
  void Parser::parseShift ()
//--------------------------
{
  IntermediateCode::OPERATION opcode;

  parseSum();

  for(;;) 
  {
    switch ( readNextToken() )
    {
      case Token::SHL : opcode = IntermediateCode::SHL; break;
      case Token::SHR : opcode = IntermediateCode::SHR; break;

      default : opcode = IntermediateCode::NOP;
    }
    if ( IntermediateCode::NOP != opcode )
    {
      readNextToken();
      parseSum();
      putCode(opcode);
    }
    else
    { 
      putBackToken();
      break;
    }
  }  
}


//------------------------
  void Parser::parseSum ()
//------------------------
{
  IntermediateCode::OPERATION opcode;

  parseProduct();

  for(;;) 
  {
    switch ( readNextToken() )
    {
      case Token::PLUS  : opcode = IntermediateCode::ADD; break;
      case Token::MINUS : opcode = IntermediateCode::SUB; break;

      default : opcode = IntermediateCode::NOP;
    }
    if ( IntermediateCode::NOP != opcode )
    {
      readNextToken();
      parseProduct();
      putCode(opcode);
    }
    else
    { 
      putBackToken();
      break;
    }
  }  
}


//----------------------------
  void Parser::parseProduct ()
//----------------------------
{
  IntermediateCode::OPERATION opcode;

  parseOperand();

  for(;;) 
  {
    switch ( readNextToken() )
    {
      case Token::MUL : opcode = IntermediateCode::MUL; break;
      case Token::DIV : opcode = IntermediateCode::DIV; break;
      case Token::MOD : opcode = IntermediateCode::MOD; break;

      default : opcode = IntermediateCode::NOP;
    }
    if ( IntermediateCode::NOP != opcode )
    {
      readNextToken();
      parseOperand();

      try
      {
        putCode(opcode);
      }
      catch (Error& err)
      {
        std::string errMsg;
        errMsg = Text::get(Text::DIV_ZERO);
        complain(errMsg);
        throw err;
      }
    }
    else
    { 
      putBackToken();
      break;
    }
  }  
}


//----------------------------
  void Parser::parseOperand ()
//----------------------------
{
  Identifier* identifier;
  Identified* identified;

  switch ( input->getToken().id  )
  {
    case Token::PLUS:  
      
      readNextToken(); 
      parseOperand(); 
      break;

    case Token::MINUS: 
      
      readNextToken(); 
      parseOperand(); 
      putCode(IntermediateCode::NEG);
      break;

    case Token::NOT: 
      
      readNextToken(); 
      parseOperand(); 
      putCode(IntermediateCode::NOT);
      break;

    case Token::LITERAL_INT:
    case Token::LITERAL_CHAR:

      putImmLoadValueCode(input->getToken().value);
      break;

    case Token::LITERAL_FLOAT:

      putImmLoadFloatValueCode(input->getFloat());
      break;

    case Token::LEFT_PARENTHESIS:

      readNextToken();
      parseTerm(); 
      readNextToken(Token::RIGHT_PARENTHESIS);
      break;

    case Token::IDENTIFIER:

      identifier = parseIdentifier();
      identified = dynamic_cast<Identified*>(identifier->getValue().get());

      if ( identified->what() == FUNCTION_IDENT )
      {
        Function* function = dynamic_cast<Function*>(identified);
        function->use();
        parseActualParameter(function);
    
        if ( function->isInline() )
          chainCodeSeq (function->getImCodeSeq());

        putCodeWithObject(IntermediateCode::CALL, function);
    
        seq.top()->requireMemoryForCall(function->getImCodeSeq()->getMemoryRequirements());
      }
      else
      {
        bool isIndirect = false;
        expect(identifier);
        Data* data = dynamic_cast<Data*>(identified);

        if ( isParsingConstTerm ) 
        {
        }
        else
        {
          data->use();
          data = parseDataAccess(data, isIndirect, true);
        }

        if ( !data->getType()->isNumericType() )
        {
          std::string errMsg = Text::get(Text::TYPE_MISMATCH);
          complain(errMsg);  
          throw Error(Error::TYPE_MISMATCH);
        }

        if ( isIndirect )
          putCodeWithObject(IntermediateCode::R_LOAD, data);
        else
          putCodeWithObject(IntermediateCode::LOAD, data);
      } 
      break;

    default: 
    {
      std::string errMsg;
      errMsg = Text::get(Text::UNEXPECTED);
      errMsg += " ";
      errMsg += Text::get(input->getToken().id);
      complain(errMsg);
      throw Error();
    }
  }
}


