/********************************************************************

    project C-Control II Compiler

    parstrng.cpp

********************************************************************/

#include "parser.h"
#include "error.h"
#include "text.h"


//-------------------------------------------------------------
  void Parser::parseStringConcatenation ( Variable* stringVar )
//-------------------------------------------------------------
{
  parseStringOperand(stringVar);

  while ( readNextToken() == Token::PLUS ) 
  {
    readNextToken();
    parseStringOperand(stringVar);
  }  

  putBackToken();
}
  
  
//-------------------------------------------------------
  void Parser::parseStringOperand ( Variable* stringVar )
//-------------------------------------------------------
{
  Identifier* identifier;
  Identified* identified;

  putCodeWithObject(IntermediateCode::LOAD_PTR, stringVar);

  switch ( input->getToken().id  )
  {
    case Token::LITERAL_STRING:

      putCodeWithString(IntermediateCode::STR_PUTISTR, input->getString());
      break;

    case Token::IDENTIFIER:

      savePosition();

      identifier = parseIdentifier();
      identified = dynamic_cast<Identified*>(identifier->getValue().get());

      if ( identified->what() > DATA_IDENT ) 
      {
        Data* data = dynamic_cast<Data*>(identified);
        data->use();

        if ( identified->what() == STRING_CONST_IDENT )
        {
          putCodeWithObject(IntermediateCode::STR_PUTCSTR, data);
          break;
        }
        else if ( identified->what() == STRING_ARRAY_CONST_IDENT )
        {
          putCodeWithObject(IntermediateCode::LOAD_PTR, data);

          if ( readNextToken() == Token::LEFT_BRACKET ) 
            parseArrayElementAccess(data);
          else
            putBackToken();

          putCodeWithObject(IntermediateCode::STR_PUTCSTR, data);
          break;
        }
        else if ( identified->what() == VAR_IDENT )
        {
          bool isIndirect = false;
          Variable* variable = dynamic_cast<Variable*>(parseDataAccess(dynamic_cast<Data*>(identified), isIndirect, false));

          if ( variable->getType()->isStringType() )
          {
            putCode(IntermediateCode::STR_PUTSTR);
            break;
          }
        }
      }

      restorePosition();

    default: 

      parseOperand();
      putCodeWithID(IntermediateCode::CAST, NumericType::INT);
      putCode(IntermediateCode::STR_PUTCH);
      break;
  }
}

