/********************************************************************

    project C-Control II Compiler

    procedure.h

********************************************************************/


#include "procedure.h"


//--- Procedure -----------------------------------------------------

//-----------------------
  Procedure::Procedure ()
//-----------------------
{ 
  identifier = 0; 
  nextLocalVarOffset = 0;
  address = 0; 
}


//------------------------
  Procedure::~Procedure ()
//------------------------
{
  int i;

  for ( i = 0; i < tempStrings.size(); i++ )
    delete tempStrings[i];
}


//--------------------------------------------------
  void Procedure::addLocalVariable ( Variable* var )
//--------------------------------------------------
{
  localVar.push_back(var);  

  if ( !var->isIndirectVariable() )
  {
    if ( var->getSize() > 1 )
    {
      nextLocalVarOffset++;
      nextLocalVarOffset &= ~1;
    }
    var->setOffset(nextLocalVarOffset);
    nextLocalVarOffset += var->getSize();
    imCodeSeq.setLocalMemSize(nextLocalVarOffset);
  }
}


//--------------------------------------------
  Variable* Procedure::getTempString ( int i )
//--------------------------------------------
{
  Variable* variable;

  if ( tempStrings.size() < (i+1) )
  {
    variable = new Variable(&StandardType::STRING, Data::LOCAL);   
    addLocalVariable(variable); 
    tempStrings.push_back(variable);  
  } 
  else
    variable = tempStrings[i];
  
  return variable;    
}



//--- Thread --------------------------------------------------------

//------------------------------------------
  Thread::Thread ( unsigned char _priority )
//------------------------------------------
: Procedure()
{ 
  id = 0; 
  priority = _priority; 
}




//--- Function ------------------------------------------------------


//---------------------------------------
  Function::Function ( bool _inlineFlag )
//---------------------------------------
: Procedure ()
{  
  inlineFlag = _inlineFlag;
  paramMemSize = 0;
  type = 0; 
  capvar = 0;
}


//-------------------------------------------
  void Function::addParameter ( Data* param )
//-------------------------------------------
{
  parameter.push_back(param);  
  paramMemSize += param->getSize();    
}


//-------------------------------------------------
  unsigned Function::getExecutionFrameSize () const
//-------------------------------------------------
{ 
  return Procedure::PTR_SIZE + Data::PTR_SIZE;
}


//----------------------------------
  void Function::locateParameters ()
//----------------------------------
{
  int offset = 0;

  for ( int i = getNumParams() - 1; i >= 0; i-- )
  {
    Data* param = parameter[i];

    offset -= param->getSize();
    param->setOffset(offset);  
  }
}



