/********************************************************************

    project C-Control II Compiler

    procedure.h

********************************************************************/

#ifndef procedure_h
#define procedure_h


#include "classlib/classlib.h"

#include "identifier.h"
#include "type.h"
#include "data.h"
#include "imcode.h"




//-----------------------------------
  class Procedure : public Identified
//-----------------------------------
{
  public:

    enum { PTR_SIZE = 2 };

    struct CodeRange { unsigned from; unsigned to; };

  protected:

    IntermediateCodeSequence imCodeSeq;

    unsigned address;
    unsigned lastaddress;
    unsigned nextLocalVarOffset;
    
    std::vector<Variable*> localVar;
    std::vector<Variable*> tempStrings;

  public:

    Procedure ();
   ~Procedure ();

    IntermediateCodeSequence* getImCodeSeq ();

    void setAddress ( unsigned _address );
    void setLastAddress ( unsigned _address );
    unsigned getAddress () const;
    unsigned getLastAddress () const;

    unsigned getLocalMemSize () const;

    void addLocalVariable ( Variable* var );

    void locateLocalVariables ();

    Variable* getTempString ( int i );
};


inline IntermediateCodeSequence* Procedure::getImCodeSeq ()
{ return &imCodeSeq; }

inline void Procedure::setAddress ( unsigned _address )
{ address = _address; }

inline void Procedure::setLastAddress ( unsigned _address )
{ lastaddress = _address; }

inline unsigned Procedure::getAddress () const
{ return address; }

inline unsigned Procedure::getLastAddress () const
{ return lastaddress; }


inline unsigned Procedure::getLocalMemSize () const
{ return nextLocalVarOffset; }



//-------------------------------
  class Thread : public Procedure
//-------------------------------
{
  protected:

    unsigned char id;
    unsigned char priority;

    Thread ();

  public:

    enum { IDLE_PRIORITY = 0, NORMAL_PRIORITY = 32 };

    Thread ( unsigned char _priority );

    void setID ( unsigned char _id );
    unsigned char getID  () const;
    unsigned char getPriority () const;
};


inline void Thread::setID ( unsigned char _id )
{ id = _id; }

inline unsigned char Thread::getID () const
{ return id; }

inline unsigned char Thread::getPriority () const
{ return priority; }



//---------------------------------
  class Function : public Procedure
//---------------------------------
{
  protected:

    unsigned  paramMemSize;
    Type*     type;
    Variable* capvar;
    bool      inlineFlag;

    std::vector<Data*> parameter;
    
  public:

    Function ( bool _inlineFlag = false );

    void setType ( Type* _type );
    Type* getType () const;

    void addParameter ( Data* param );
    Data* getParameter ( int i );
    unsigned getNumParams () const;
    unsigned getParamMemSize () const;
    unsigned getExecutionFrameSize () const;

    void setCapVar ( Variable* _capvar );
    Variable* getCapVar ();

    void locateParameters ();

    bool isInline ();
};





inline void Function::setType ( Type* _type )
{ type = _type; }

inline Type* Function::getType () const
{ return type; }

inline Data* Function::getParameter ( int i )
{ return parameter[i]; }

inline unsigned Function::getNumParams () const 
{ return parameter.size(); }

inline unsigned Function::getParamMemSize () const
{ return paramMemSize; }

inline void Function::setCapVar ( Variable* _capvar )
{ capvar = _capvar; }

inline Variable* Function::getCapVar ()
{ return capvar; }

inline bool Function::isInline ()
{ return inlineFlag; }

#endif