/********************************************************************

    project C-Control II Compiler

    program.cpp

********************************************************************/

#include "program.h"
#include "text.h"


//-------------------
  Program::Program ()
//-------------------
: root("",Virtual::Pointer())
{
  resetScope();
  instructionNo = 0;
  shallLogInstructions = false;
  isGood = false;
}

 
//----------------------------------------------------------
  long Program::logInstruction ( long hModule, long lineNo )
//----------------------------------------------------------
{
  long number = 0;

  if ( shallLogInstructions )
  {
    if ( instructionNo == 65535 )
      return -1;

    number = instructionNo++;

//    if ( instructionLogs.capacity() < instructionNo )
//      instructionLogs.reserve(number + 1024);
  
    INSTLOG log = { hModule,lineNo };
    instructionLogs.push_back(log);
//    instructionLogs[number].hModule = hModule;
//    instructionLogs[number].lineNo = lineNo;
  }

  return number;
}


//----------------------------------------------------------
  unsigned short Program::logUsedInstruction ( long instNo )
//----------------------------------------------------------
{
  if ( instNo == lastInstructionNo )
    return lastUsedInstructionNo;

  if ( -1 == instNo )
    return (unsigned short) -1;

  std::map<long,long>::iterator iterator = mapOfInstNumbersToUsedNumbers.find(instNo);
  
  if ( iterator != mapOfInstNumbersToUsedNumbers.end()  )
  {
    lastInstructionNo = instNo;
    lastUsedInstructionNo = (*iterator).second;
    return lastUsedInstructionNo;
  }

  lastInstructionNo = instNo;
    
  unsigned short number = usedInstructionNo++;

  INSTLOG log = { instructionLogs[instNo].hModule, instructionLogs[instNo].lineNo };
  usedInstructionLogs.push_back(log);

  lastUsedInstructionNo = number;
  mapOfInstNumbersToUsedNumbers[instNo] = number;
  
  return number;
}




//----------------------------------------------------------------------------------
  Identifier* Program::declare ( unsigned long lino, const std::string& name, 
                                 Identified* object, IDENTCLASS idclass, bool hide )
//----------------------------------------------------------------------------------
{
   Identifier* identifier = scope.top()->addChild(name, Virtual::Pointer(object));

  if ( identifier )
  {
    object->init(lino, identifier, idclass);
    identifier->hideChildren(hide);
    return identifier;
  }
  else
    return 0;
}


//-----------------------------------------------------
  void Program::restoreScope ( Identifier* identifier )
//-----------------------------------------------------
{
  while ( scope.top() != identifier )
    scope.pop();
}


//---------------------------
  void Program::resetScope ()
//---------------------------
{ 
  scope.clear();
  scope.push(&root);
  isGood = false;
}


//---------------------------
  void Program::beginBuild ()
//---------------------------
{
  // get ready for next build
  instructionNo = 0;
  instructionLogs.clear();
  instructionLogs.reserve(8192);
  shallLogInstructions = true;
  isGood = false;
}


//---------------------------------------------------
  bool Program::finishBuild ( std::ostream* _errLog )
//---------------------------------------------------
{
  bool success = true;

  try
  {
    __finishBuild ();
  }

  catch ( BUILD_ERROR error )
  {
    switch ( error )
    {
      case ERR_NO_MAIN_THREAD : (*_errLog) << Text::get(Text::NO_MAIN_THREAD); break;
      case ERR_TOO_MANY_THREADS : (*_errLog) << Text::get(Text::TOO_MANY_THREADS); break;
      case ERR_TOO_MUCH_DATA : (*_errLog) << Text::get(Text::TOO_MUCH_VARIABLE_DATA); break;
      case ERR_TOO_MANY_CONSTANTS : (*_errLog) << Text::get(Text::TOO_MUCH_CONSTANT_DATA); break;
      case ERR_TOO_MUCH_CODE : (*_errLog) << Text::get(Text::PROGRAM_TOO_BIG); break;
    }
    success = false;
  }

  catch (...)
  {
    success = false;
   (*_errLog) << Text::get(Text::FATAL_ERROR); 
  }

  if ( success )
  {
    (*_errLog) << Text::get(Text::VMC_READY); 
    isGood = true;
  }

  (*_errLog) << std::endl << '\x0';

  shallLogInstructions = false;
  return success;
}


//------------------------------
  void Program::__finishBuild ()
//------------------------------
{
//  std::ofstream os("imcode.txt");

  int i, k;
  unsigned id = 1;
  bool main_thread_found = false;

  std::vector<Function*> functions;
  std::vector<Thread*> threads;

  functions.reserve(100);   // xxx test lower values!!!
  threads.reserve(20);

  Identifier* moduleIdentifier;
  Identifier* identifier;


  for ( i = 0, moduleIdentifier = root.getFirstChild(); 
        i < root.getNumChildren(); 
        i++, moduleIdentifier = root.getNextChild() )
  {
    for ( k = 0, identifier = moduleIdentifier->getFirstChild(); 
          k < moduleIdentifier->getNumChildren(); 
          k++, identifier = moduleIdentifier->getNextChild() )
    {
      Identified* identified = dynamic_cast<Identified*> (identifier->getValue().get());

      Function* function;
      Thread* thread;

      switch ( identified->what() )
      {
        case FUNCTION_IDENT : 

          function = dynamic_cast<Function*>(identified);
          if ( function->isUsed() )
            functions.push_back(function);
          break;

        case THREAD_IDENT   : 

          if ( id > 255 ) throw(ERR_TOO_MANY_THREADS);
          thread = dynamic_cast<Thread*>(identified);
          threads.push_back(thread);
          thread->setID(id++);
          if ( thread->getPriority() ) main_thread_found = true;
          break;
      }
    }
  }

  if ( !main_thread_found ) throw(ERR_NO_MAIN_THREAD);

  backend.init(threads.size());

  usedDataLogs.clear();
  usedDataLogs.reserve(100);   

  for ( i = 0, moduleIdentifier = root.getFirstChild(); 
        i < root.getNumChildren(); 
        i++, moduleIdentifier = root.getNextChild() )
  {
    for ( k = 0, identifier = moduleIdentifier->getFirstChild(); 
          k < moduleIdentifier->getNumChildren(); 
          k++, identifier = moduleIdentifier->getNextChild() )
    {
      Identified* identified = dynamic_cast<Identified*> (identifier->getValue().get());

      if ( !identified->isUsed() ) 
        continue;

      Variable* variable;
      Constant* constant;
      ArrayConstant* arrayconst;
      StringConstant* stringconst;
      StringArrayConstant* stringarrayconst;

      switch ( identified->what() )
      {
        case VAR_IDENT :

          variable = dynamic_cast<Variable*>(identified);
          backend.allocateVariable(variable);
          usedDataLogs.push_back(dynamic_cast<Data*>(identified));
          break;

        case CONST_IDENT :

          constant = dynamic_cast<Constant*>(identified);
          constant->setOffset(backend.storeConst(constant));
          usedDataLogs.push_back(dynamic_cast<Data*>(identified));
          break;

        case ARRAY_CONST_IDENT :

          arrayconst = dynamic_cast<ArrayConstant*>(identified);
          arrayconst->setOffset(backend.storeConst(arrayconst));
          break;

        case STRING_CONST_IDENT :

          stringconst = dynamic_cast<StringConstant*>(identified);
          stringconst->setOffset(backend.storeConst(stringconst));
          break;

        case STRING_ARRAY_CONST_IDENT :
         
          stringarrayconst = dynamic_cast<StringArrayConstant*>(identified);
          stringarrayconst->setOffset(backend.storeConst(stringarrayconst));
          break;
      }

      if ( backend.getNumConstantBytes() > backend.getConstantMemorySize() )
        throw ERR_TOO_MANY_CONSTANTS;
         
      if ( backend.getNumVariableBytes() > backend.getDataMemorySize() )
        throw ERR_TOO_MUCH_DATA;
    }
  }  

//  os << "\n\n* threads\n";

  for ( i = 0; i < threads.size(); i++ )
  {
    Thread* thread = threads[i];
    std::string name;
    IntermediateCodeSequence* imseq = thread->getImCodeSeq();
    getFullName(thread->getIdentifier(), name);
    unsigned short codeOffset = (unsigned short) backend.compile(imseq, name);
    backend.setThreadInitData(i, codeOffset, imseq->getMemoryRequirements(), 
                              thread->getLocalMemSize(), thread->getPriority());
    thread->setAddress(codeOffset);
    thread->setLastAddress(backend.getCodeOffset()-1);

    if ( backend.getNumVariableBytes() > backend.getDataMemorySize() )
      throw ERR_TOO_MUCH_DATA;

    if ( backend.getNumCodeWords() > backend.getCodeMemorySize() )
      throw ERR_TOO_MUCH_CODE;

//    os << '\n' << name.c_str() << '\n';
//    imseq->dump(os);

    for ( k = 0, identifier = thread->getIdentifier()->getFirstChild(); 
          k < thread->getIdentifier()->getNumChildren(); 
          k++, identifier = thread->getIdentifier()->getNextChild() )
    {
      Identified* identified = dynamic_cast<Identified*> (identifier->getValue().get());

      if ( (identified->what() == VAR_IDENT) && identified->isUsed() )
         usedDataLogs.push_back(dynamic_cast<Data*>(identified));
    }
  }

//  os << "\n\n* functions\n";

  for ( i = 0; i < functions.size(); i++ )
  {
    Function* func = functions[i];
    std::string name;
    IntermediateCodeSequence* imseq = func->getImCodeSeq();
    getFullName(func->getIdentifier(), name);

    func->setAddress(backend.getCodeOffset());
    backend.compile(imseq, name);
    func->setLastAddress(backend.getCodeOffset()-1);

    if ( backend.getNumCodeWords() > backend.getCodeMemorySize() )
      throw ERR_TOO_MUCH_CODE;

//    os << '\n' << name.c_str() << '\n';
//    imseq->dump(os);

    for ( k = 0, identifier = func->getIdentifier()->getFirstChild(); 
          k < func->getIdentifier()->getNumChildren(); 
          k++, identifier = func->getIdentifier()->getNextChild() )
    {
      Identified* identified = dynamic_cast<Identified*> (identifier->getValue().get());

      if ( (identified->what() == VAR_IDENT) && identified->isUsed() )
         usedDataLogs.push_back(dynamic_cast<Data*>(identified));
    }
  }

  backend.finish();

  usedInstructionNo = 0;
  lastInstructionNo = -1;
  lastUsedInstructionNo = -1;

  usedInstructionLogs.clear();
  usedInstructionLogs.reserve(8192);
  mapOfInstNumbersToUsedNumbers.clear();
  mapOfInstNumbersToUsedNumbers[-1] = (unsigned short)-1;

  for ( i = 0; i < backend.getNumCodeWords(); i++ )
    logUsedInstruction(backend.getCodeInstructionNo(i));
}


//--------------------------------------------------
  bool Program::writeOutputFile ( std::ostream& os )
//--------------------------------------------------
{
  if ( !isGood ) return false;   

  long i;
  int value;

  long numStoredConstBytes = (backend.getNumConstantBytes() + 31) & 0xFFFFFFE0;
  long numStoredCodeWords = (backend.getNumCodeWords() + 15) & 0xFFFFFFF0;

  os <<  "CC2VMC" << "\n"; 
 
  os << numStoredConstBytes << '\n' << numStoredCodeWords << "\n";
  
  // *** write constant bytes, fill with FF to multiples of 32 ***

  for ( i = 0; i < backend.getNumConstantBytes(); ++i )
  {
    if ( !( i & 0x1F ) ) os << '\n';
    value = backend.getConstByte(i);
    os << value << ' ';
  }
  for ( ; i < numStoredConstBytes; ++i )
  {
    if ( !( i & 0x1F ) ) os << '\n';
    os << 0xFF << ' ';
  }
  os << '\n';

  
  // *** write code words, fill with FFFF to multiples of 16 ***

  for ( i = 0; i < backend.getNumCodeWords(); ++i )
  {
    if ( !(i & 0x0F) ) os << '\n';

    value = backend.getCodeWord(i);
    os <<  value << ' ';
  }
  for ( ; i < numStoredCodeWords; ++i )
  {
    if ( !(i & 0x0F) )  os << '\n';
    os << 0xFFFF  << ' ';
  }
  os << '\n';

  return !os.fail();
}


//-----------------------------------------------------------------------
  bool Program::writeOutputFile ( std::ostream& os, std::ostream& dbgos )
//-----------------------------------------------------------------------
{
  if ( !isGood ) return false;   

  long i;
  int value;

  long numStoredConstBytes = (backend.getNumConstantBytes() + 31) & 0xFFFFFFE0;
  long numStoredCodeWords = (backend.getNumCodeWords() + 15) & 0xFFFFFFF0;

  os <<  "CC2VMC" << "\n"; 
 
  os << numStoredConstBytes << '\n' << numStoredCodeWords << "\n";
  dbgos << numStoredConstBytes << '\n' << numStoredCodeWords << "\n";

  
  // *** write constant bytes, fill with FF to multiples of 32 ***

  for ( i = 0; i < backend.getNumConstantBytes(); ++i )
  {
    if ( !( i & 0x1F ) ) 
    {
      os << '\n';
      dbgos << '\n';
    }
    value = backend.getConstByte(i);
    os << value << ' ';
    dbgos << value << ' ';
  }
  for ( ; i < numStoredConstBytes; ++i )
  {
    if ( !( i & 0x1F ) ) 
    {
      os << '\n';
      dbgos << '\n';
    }
    os << 0xFF << ' ';
    dbgos << 0xFF << ' ';
  }
  os << '\n';
  dbgos << '\n';

  
  // *** write code words, fill with FFFF to multiples of 16 ***

  for ( i = 0; i < backend.getNumCodeWords(); ++i )
  {
    if ( !(i & 0x0F) )
    {
      os << '\n';
      dbgos << '\n';
    }

    dbgos << mapOfInstNumbersToUsedNumbers[backend.getCodeInstructionNo(i)] << ' ';

    value = backend.getCodeWord(i);
    os <<  value << ' ';
    dbgos <<  value << ' ';
  }
  for ( ; i < numStoredCodeWords; ++i )
  {
    if ( !(i & 0x0F) )
    {
      os << '\n';
      dbgos << '\n';
    }

    dbgos << 0xFFFF  << ' ';
  
    os << 0xFFFF  << ' ';
    dbgos << 0xFFFF  << ' ';
  }
  os << '\n';
  dbgos << '\n';


  // *** list used and logged instruction/data information (debug info) ***

  // instruction debug info

  dbgos << '\n';
  dbgos << usedInstructionNo;

  for ( i = 0; i < usedInstructionNo; ++i )
  {
    if ( !(i & 0x0F) ) dbgos << '\n';
    
    dbgos << usedInstructionLogs[i].hModule << ' ' << usedInstructionLogs[i].lineNo << ' ';     
  }
  dbgos << "\n\n" ;

  
  // data debug info
   
  dbgos << usedDataLogs.size();
  dbgos << '\n';

  for ( i = 0; i < usedDataLogs.size(); ++i )
  {
    Data* data = usedDataLogs[i];
    Type* type = data->getType();

    if ( 
         type->isUserType () || 
         data->getNumArrayElements() || 
         (data->getMemoryClass() == Data::CONSTANT)
       ) 
      continue;

    Identifier* ident =  data->getIdentifier();

    // name
    ident->writePath(dbgos, '.');  
    dbgos << ' ';

    // memory class
    dbgos << data->getMemoryClass() << ' ';

    // address 
    Data* reference = dynamic_cast<Variable*>(data)->getReference();
    if ( reference )
      dbgos << reference->getOffset() - 4 << ' ';  // dirty: -4 means before bp and sp on stack
    else
      dbgos << (data->getOffset() < 0? data->getOffset() - 4 : data->getOffset()) << ' ';

    // type
    if ( type->isStringType () )
      dbgos << -1;
    else 
      dbgos << dynamic_cast<NumericType*>(type)->getForm();

    // valid scope
    if ( data->getMemoryClass() == Data::LOCAL )
    {
      Procedure* proc = dynamic_cast<Procedure*>(ident->getParent()->getValue().get());
      dbgos << ' ' << proc->getAddress() << ' ' << proc->getLastAddress();
    }
    else
      dbgos << ' ' << 0 << ' ' << 0;

    dbgos << '\n';
  }

  return !( os.fail() || dbgos.fail() );
}


//-----------------------------------------
  long Program::getRequiredVMCBufferSize ()
//-----------------------------------------
{
  if ( !isGood ) return 0;   

  long numStoredConstBytes = (backend.getNumConstantBytes() + 31) & 0xFFFFFFE0;
  long numStoredCodeWords = (backend.getNumCodeWords() + 15) & 0xFFFFFFF0;
  
  return 8 + numStoredConstBytes + 2 * numStoredCodeWords;
}


//-----------------------------------------------------
  void Program::fillVMCBuffer ( unsigned char* buffer )
//-----------------------------------------------------
{
  if ( !isGood ) return;   

  long storepos, i;
  long numStoredConstBytes = (backend.getNumConstantBytes() + 31) & 0xFFFFFFE0;
  long numStoredCodeWords = (backend.getNumCodeWords() + 15) & 0xFFFFFFF0;

  buffer[0] = (unsigned char)  numStoredConstBytes;
  buffer[1] = (unsigned char) (numStoredConstBytes >> 8);
  buffer[2] = (unsigned char) (numStoredConstBytes >> 16);
  buffer[3] = (unsigned char) (numStoredConstBytes >> 24);
  buffer[4] = (unsigned char)  numStoredCodeWords;
  buffer[5] = (unsigned char) (numStoredCodeWords >> 8);
  buffer[6] = (unsigned char) (numStoredCodeWords >> 16);
  buffer[7] = (unsigned char) (numStoredCodeWords >> 24);

  storepos = 8;

  for ( i = 0; i < backend.getNumConstantBytes(); i++ ) 
    buffer[storepos++] = backend.getConstByte(i);   

  for ( ; i < numStoredConstBytes; i++ ) 
    buffer[storepos++] = (unsigned char) 0xFF;

  for ( i = 0; i < backend.getNumCodeWords(); i++ ) 
  {
    unsigned short codeword = backend.getCodeWord(i);
    buffer[storepos++] = (unsigned char) (codeword & 0xFF);   
    buffer[storepos++] = (unsigned char) (codeword >> 8);   
  }

  for ( ; i < numStoredCodeWords; i++ ) 
  {
    buffer[storepos++] = (char) 0xFF;   
    buffer[storepos++] = (char) 0xFF;   
  }
}

