/********************************************************************

    project C-Control II Compiler

    program.h

********************************************************************/

#ifndef program_h
#define program_h

#include "classlib/classlib.h"

#include "identifier.h"

#include "type.h"
#include "data.h"
#include "procedure.h"
#include "imcode.h"
#include "backend.h"


typedef enum
{
  MODULE_IDENT, TYPE_IDENT, FIELD_IDENT, FUNCTION_IDENT, THREAD_IDENT,
  DATA_IDENT, VAR_IDENT, CONST_IDENT, ARRAY_CONST_IDENT, STRING_CONST_IDENT, STRING_ARRAY_CONST_IDENT

} IDENTCLASS;

//-------------
  class Program
//-------------
{
  protected:

    BackEnd    backend;    
    Identifier root;
    bool       shallLogInstructions;
    bool       isGood;
  
    struct _scope_ : public std::stack<Identifier*> {void clear(){c.clear();}} scope; 

    struct INSTLOG 
    { 
      long hModule; 
      long lineNo; 
    };

    long           instructionNo;
    long           lastInstructionNo;
    unsigned short usedInstructionNo;
    unsigned short lastUsedInstructionNo;
    
    std::vector<INSTLOG> instructionLogs;
    std::vector<INSTLOG> usedInstructionLogs;
    std::vector<Data*>   usedDataLogs;
    std::map<long, long> mapOfInstNumbersToUsedNumbers;

    unsigned short logUsedInstruction ( long instNo );

    typedef enum { ERR_NO_MAIN_THREAD, ERR_TOO_MANY_THREADS, ERR_TOO_MUCH_DATA, 
                   ERR_TOO_MANY_CONSTANTS, ERR_TOO_MUCH_CODE } BUILD_ERROR;

  public:

    Program ();

    long logInstruction ( long hModule, long lineNo ); 
    long getLoggedInstructionNo ();

    Identifier* declare ( unsigned long lino, const std::string& name, 
                          Identified* object, IDENTCLASS idclass, bool hide = false );

    void beginBuild ();
    void __finishBuild ();
    bool finishBuild ( std::ostream* _errLog = 0 );
    void cancelInstructionLogging ();

    bool writeOutputFile ( std::ostream& os );
    bool writeOutputFile ( std::ostream& os, std::ostream& dbgos );
    long getRequiredVMCBufferSize ();
    void fillVMCBuffer ( unsigned char* buffer );

    Identifier* getRootScope ();
    Identifier* getScope ();
    void enterScope ( Identifier* pScope );
    void leaveScope ();
    void restoreScope ( Identifier* pScope );
    void resetScope ();
};


inline Identifier* Program::getRootScope ()
{ return &root; }

inline Identifier* Program::getScope ()
{ return scope.top(); }

inline void Program::enterScope ( Identifier* pIdent )
{ scope.push(pIdent); }

inline void Program::leaveScope ()
{ scope.pop(); }

inline long Program::getLoggedInstructionNo ()
{ return instructionNo - 1; }


inline void Program::cancelInstructionLogging ()
{ shallLogInstructions = false; }

/*
inline Program::ExUnknownIdent::ExUnknownIdent ( const std::string& _name )
{ name = _name;}


inline void Program::requireMemoryForCall ( int size )
{ seq.top()->requireMemoryForCall(size); }

inline unsigned Program::getMark ()
{ return seq.top()->getMark(); }

inline unsigned Program::mark ()
{ return seq.top()->mark(); }

inline void Program::mark ( unsigned id )
{ seq.top()->mark(id); }

*/

#endif
