/********************************************************************

    project C-Control II Compiler

    project.h

********************************************************************/

#ifndef project_h
#define project_h

#include "classlib/classlib.h"


#include "module.h"
#include "program.h"
#include "scanner.h"
#include "parser.h"
#include "backend.h"

//-------------
  class Project
//-------------
{
  public:

    typedef std::list<Module*>                     ModuleList;
    typedef ModuleList::iterator                   ModuleIterator;
    typedef unsigned long                          ModuleHandle;
    typedef std::map<ModuleHandle, ModuleIterator> ModuleHandleMap;

    enum { ERR_BAD_MODULE_NAME = -2, ERR_DUPLICATED_MODULE_NAME = -1, FIRST_FREE_MODULE_HANDLE = 1 };
    
  protected:

    std::string     name;
    ModuleHandle    nextFreeModuleHandle;
    ModuleList      moduleList;
    ModuleHandleMap moduleHandleMap;
    Program         program;
    Scanner         scanner;
    Parser          parser;

    std::string     errLogFileName;
    char*           errLogBuf;
    int             errLogBufLength;
    std::ostream*   errLog;

    Project ();

    bool hasModule ( const std::string& modulename );
    long compile ( ModuleIterator last );
    long compile ( Module* module );

    bool openErrLog ();

  public:

    Project ( const std::string& _name );
   ~Project ();
    
    void rename ( const std::string& _name );

    bool isGoodName ( const std::string& modulename );
    ModuleHandle add ( const std::string& modulename, const std::string& filename, char* strbuf = 0 );

    Module* getModule ( ModuleHandle hModule );

    void removeAll ();   
    bool remove ( ModuleHandle hModule );   
    bool move ( ModuleHandle hModule, ModuleHandle hOther );
    long compile ( ModuleHandle hModule );
    long build ();
    void beginBuild ();
    bool finishBuild ();
    bool writeOutputFile ( const std::string& filename, bool doWriteDebugFile = true );
    long getRequiredVMCBufferSize ();
    void fillVMCBuffer ( unsigned char* buffer );

    void setErrLog ( const std::string& filename ); 
    void setErrLog ( char* buf, int length );

    Identifier* getIdentifierInContext ( const char* ident_c_str, ModuleHandle hModule, unsigned long lino );
    bool getIdentDefPos ( const char* ident_c_str, ModuleHandle* hModule_ptr, unsigned long* lino_ptr );
    bool getIdentCompletions ( const char* ident_c_str, ModuleHandle hModule, unsigned long lino, 
                               char* completions, unsigned long length );
};


#endif