/********************************************************************

    project C-Control II Compiler

    scanner.h

********************************************************************/

#ifndef scanner_h
#define scanner_h

#include "classlib/classlib.h"
#include "token.h"


//-----------------
  class ScanResult
//-----------------
{
  protected:

    Token*       tokens;
    double*      floats;
    std::string* strings;

    unsigned long tokenCapacity;
    unsigned long floatCapacity;
    unsigned long stringCapacity;

    unsigned long numberOfTokens;
    unsigned long numberOfFloats;
    unsigned long numberOfStrings;

    unsigned long iToken;
    unsigned long iFloat;
    unsigned long iString;

  public:

    ScanResult ();
   ~ScanResult ();
   
    void init ();
    void free ();

    void putToken ( Token::ID id );
    void putToken ( Token::ID id, long value );
    void putFloatToken ( Token::ID id, double value );
    void putStringToken ( Token::ID id, std::string const& string );

    void finish ();

    unsigned long getNumberOfTokens () const;
    unsigned long getPosition () const;

    unsigned long gotoToken ( unsigned long _iToken = 0 );
    unsigned long gotoNextToken ();
    unsigned long putBackToken ();
    
    Token const&  getToken () const;
    std::string& getString () const;
    std::string& getString ( Token& token ) const;
    double getFloat () const;
    double getFloat ( Token& token ) const;
};


inline unsigned long ScanResult::getNumberOfTokens () const 
{ return numberOfTokens; }

inline unsigned long ScanResult::getPosition () const 
{ return iToken; }

inline unsigned long ScanResult::gotoToken ( unsigned long _iToken ) 
{ return iToken = _iToken; }

inline unsigned long ScanResult::gotoNextToken () 
{ return ++iToken; }
   
inline unsigned long ScanResult::putBackToken () 
{ return --iToken; }

inline Token const& ScanResult::getToken () const
{ return tokens[iToken]; }

inline std::string& ScanResult::getString () const 
{ return strings[getToken().value]; }

inline std::string& ScanResult::getString ( Token& token ) const 
{ return strings[token.value]; }

inline double ScanResult::getFloat () const 
{ return floats[getToken().value]; }

inline double ScanResult::getFloat ( Token& token ) const 
{ return floats[token.value]; }




//-------------
  class Scanner
//-------------
{
  protected:

    unsigned char ch;
    std::istream* input;
    ScanResult*   output;

    static unsigned short hash ( const std::string& string );

    enum { KEYWORD_HASH_TABLE_LENGTH = 277 };
  
    unsigned short maxKeywordLength;
    unsigned short hashtable[KEYWORD_HASH_TABLE_LENGTH];
    
    struct  KEYWORD { Token::ID id; std::string string; };
    KEYWORD keywords[Token::KEYWORDS_END - Token::KEYWORDS_BEGIN - 1];

    Token::ID Scanner::getKeywordID ( const std::string& string );

    unsigned char getChar ();
    unsigned char getControlChar ();
    void putBackChar ();

    bool isChar ( const unsigned char cc );
    bool isNextChar ( const unsigned char cc );
    bool isNextCharI ( const unsigned char cc );
    bool isNewLine ();

    void skipWhites ();
    void invalidateChar ();

    void scanComment ();
    void scanCommentUntilEol ();

    void scanKeywordOrIdentifier ();

    void scanOperator ();
    Token::ID getDotOrEllipse ();

    void scanLiteralValue ();
    void scanLiteralChar ();
    void scanLiteralString ();

    void scanDecValue ();
    int  scanAndGetDecInteger ();
    int  scanAndGetHexInteger ();
    int  scanAndGetBinInteger ();

  public:

    Scanner ();

    void run ( std::istream* _input, ScanResult* _output );
};



inline unsigned char Scanner::getChar ()
{
  return ( ch = input->good() ? input->get() : 0 );
}

inline void Scanner::putBackChar ()
{ input->putback(ch); }

inline bool Scanner::isChar ( const unsigned char cc )
{ return ( cc == ch ); }


#endif