/********************************************************************

    project C-Control II Compiler

    types.cpp

********************************************************************/
    
#include "type.h"
    

//----------------------------------------------------------------
  NumericType::FORM NumericType::getFormFromValue ( double value )
//----------------------------------------------------------------
{
  if (  value != (long) value )
    return FLOAT;
  else if ( value != (short) value ) 
    return LONG;
  else if ( value != (unsigned char) value )
    return INT;
  else 
    return BYTE;
}


//---------------------------------------
  NumericType::NumericType ( FORM _form )
//---------------------------------------
{
  form = _form;
  cls = NUMERIC;
}


//---------------------------------------------
  void NumericType::updateForm ( double value )
//---------------------------------------------
{
  FORM _form = getFormFromValue(value);

  if ( _form > form )
    form = _form;
}
 
  
//-----------------------------------------
  NumericType::FORM NumericType::getForm ()
//-----------------------------------------
{
  return form;
}
    
    
//--------------------------------
  unsigned NumericType::getSize ()
//--------------------------------
{
  switch ( form )
  {
    case BYTE  : return 1;
    case INT   : return 2;
    case LONG  : return 4;
    case FLOAT : return 8;
    
    default : return 0;
  }  
}



//--- StringType ----------------------------------------------------


//-------------------------
  StringType::StringType ()
//-------------------------
{
  capacity = 30;
  cls = STRING;
}


//-----------------------------------
  unsigned StringType::getCapacity ()
//-----------------------------------
{
  return capacity;
}


//-------------------------------
  unsigned StringType::getSize ()
//-------------------------------
{
  // WORD length + BYTE[capacity];
  return 32; 
}


//--- UserType ------------------------------------------------------


//------------------------------------------------------------------------------------
  UserType::Field::Field ( Type* _type, unsigned _offset, unsigned _numArrayElements )
//------------------------------------------------------------------------------------
{
  type             = _type;
  offset           = _offset;
  numArrayElements = _numArrayElements;
}


//---------------------------------
  Type* UserType::Field::getType ()
//---------------------------------
{
  return type;
}


//--------------------------------------
  unsigned UserType::Field::getOffset ()
//--------------------------------------
{
  return offset;
}


//------------------------------------------------
  unsigned UserType::Field::getNumArrayElements ()
//------------------------------------------------
{
  return numArrayElements;
}


//--------------------------------
  bool UserType::Field::isArray ()
//--------------------------------
{
  return numArrayElements != 0;
}


//---------------------
  UserType::UserType ()
//---------------------
{
  nextFieldOffset = 0;
  cls = USER;
}


//-----------------------------------------------------------------------------
  UserType::Field* UserType::newField ( Type* type, unsigned numArrayElements )
//-----------------------------------------------------------------------------
{
  if ( type->getSize() > 1 )
  {
    nextFieldOffset++;
    nextFieldOffset &= ~1;
  }

  Field* field = new Field(type, nextFieldOffset, numArrayElements);

  nextFieldOffset += numArrayElements? numArrayElements * type->getSize() : type->getSize();
  
  return field;
}

   
//-----------------------------
  unsigned UserType::getSize ()
//-----------------------------
{
  return (nextFieldOffset + 1) & ~1;
}


namespace StandardType
{

NumericType BYTE(NumericType::BYTE);
NumericType INT(NumericType::INT);
NumericType LONG(NumericType::LONG);
NumericType FLOAT(NumericType::FLOAT);
StringType STRING;


//------------------------------------------
  NumericType* getFromValue ( double value )
//------------------------------------------
{
  switch ( NumericType::getFormFromValue(value) )
  {
    case NumericType::BYTE: return &BYTE;
    case NumericType::INT: return &INT;
    case NumericType::LONG: return &LONG;

    default: return &FLOAT;
  }
}

}


