/********************************************************************

    project C-Control II Compiler

    type.h

********************************************************************/

#ifndef type_h
#define type_h

#include "classlib/classlib.h"
#include "identifier.h"


class Type;
class NumericType;
class StringType;
class UserType;

//------------------------------
  class Type : public Identified
//------------------------------
{
  public:

    typedef enum { NUMERIC, STRING, USER } CLASS;

  protected:

    CLASS cls;

  public:

    virtual unsigned getSize () = 0;

    CLASS getClass () const;
    bool isNumericType () const;
    bool isStringType () const;
    bool isUserType () const;
};


inline Type::CLASS Type::getClass () const
{ return cls; }

inline bool Type::isNumericType () const 
{ return cls == NUMERIC; }

inline bool Type::isStringType () const
{ return cls == STRING; }

inline bool Type::isUserType () const
{ return cls == USER; }



//-------------------------------
  class NumericType : public Type  
//-------------------------------
{
  public:

    typedef enum { BYTE, INT, LONG, FLOAT } FORM;

  protected:

    FORM form;

  public :

    static FORM getFormFromValue ( double value );

    NumericType ( FORM _form = BYTE );

    void updateForm ( double value );
    FORM getForm ();

    virtual unsigned getSize ();
};



//------------------------------
  class StringType : public Type
//------------------------------
{
  protected:

    unsigned capacity;

  public:

    StringType ();

    virtual unsigned getSize ();
    unsigned getCapacity ();
};


//----------------------------
  class UserType : public Type
//----------------------------
{
  public:

    class Field : public Identified
    {
      friend class UserType;

      protected:

        Type*    type;
        unsigned offset;
        unsigned numArrayElements;          

        Field ();
        Field ( Type* _type, unsigned _offset, unsigned _numArrayElements );

      public:

        Type*    getType ();
        unsigned getOffset ();
        unsigned getNumArrayElements ();          
        bool     isArray ();
    };


  protected:
    
    unsigned nextFieldOffset;

  public:

    UserType ();

    Field* newField ( Type* type, unsigned numArrayElements = 0 );

    virtual unsigned getSize ();
};



namespace StandardType
{
  NumericType* getFromValue ( double value );

  extern NumericType BYTE;
  extern NumericType INT;
  extern NumericType LONG;
  extern NumericType FLOAT;
  extern StringType  STRING;
}


#endif